# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from aiootp.commons.namespaces import OpenNamespace


slick256_test_vector_0 = OpenNamespace(
    plaintext=(
        b"it is not a doctrine to be preached, but a deed to be done."
    ),
    key=bytes.fromhex(
        "54868dc506d99611adb67f1b41eda44fd7151e135860a6d791cad6df6715bd"
        "5204713bc2b064e88a429b93aaa282d9cc7c0c5cdaaffb65cc268b7acedd5e"
        "531a"
    ),
    shmac=bytes.fromhex("5e6011b3b1ae4314b4192a1e6a18b2ca8130bc5cafcd7287"),
    salt=bytes.fromhex("cf3761d57860bdbc"),
    iv=bytes.fromhex("dda31acd46832df9"),
    aad=b"test_vector_I",
    ciphertext=bytes.fromhex(
        "f73e7361a73af3eeabc88247555e06fc5bdaea10482c8351eab3ee9cfcc6fe"
        "f678b1171e6458fcef3d27896bee262769a742f8ffce51cd44db2a8e5673f4"
        "5770df4b907683c881049f15a92bed514613378ec4beb07766e809b5f68e1e"
        "326bc7"
    ),
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["identifiers"]


import keyword
from string import ascii_letters, digits
from hypothesis.strategies import DrawFn, composite, text


@composite
def identifiers(draw: DrawFn) -> str:
    first_chars = f"_{ascii_letters}"
    first_chars_st = text(alphabet=first_chars, min_size=1, max_size=1)

    final_chars = f"_{ascii_letters}{digits}"
    final_chars_st = text(alphabet=final_chars, min_size=0, max_size=16)

    result = draw(first_chars_st) + draw(final_chars_st)
    while keyword.iskeyword(result):
        result += draw(final_chars_st)

    return result



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import io
from collections import deque

from conftest import *

from aiootp import _gentools as gentools


class TestInspectTools:
    async def test_is_async_iterable(self) -> None:
        async_iterable = gentools.aunpack("test")
        iterable = gentools.unpack("test")

        assert gentools.is_async_iterable(async_iterable)
        assert not gentools.is_async_iterable(iterable)

    async def test_is_async_iterator(self) -> None:
        async_iterator = gentools.aunpack("test").__aiter__()
        iterator = gentools.unpack("test").__iter__()

        assert gentools.is_async_iterator(async_iterator)
        assert not gentools.is_async_iterator(iterator)

    async def test_is_async_generator(self) -> None:
        async_generator = gentools.aunpack("test")
        generator = gentools.unpack("test")

        assert gentools.is_async_generator(async_generator)
        assert not gentools.is_async_generator(generator)

    async def test_is_async_gen_function(self) -> None:
        async_generator_function = gentools.aunpack
        generator_function = gentools.unpack

        assert gentools.is_async_gen_function(async_generator_function)
        assert not gentools.is_async_gen_function(generator_function)

    async def test_is_generator(self) -> None:
        async_generator = gentools.aunpack("test")
        generator = gentools.unpack("test")

        assert not gentools.is_generator(async_generator)
        assert gentools.is_generator(generator)

    async def test_is_generator_function(self) -> None:
        async_generator_function = gentools.aunpack
        generator_function = gentools.unpack

        assert not gentools.is_generator_function(async_generator_function)
        assert gentools.is_generator_function(generator_function)

    async def test_is_iterable(self) -> None:
        async_iterable = gentools.aunpack("test")
        iterable = gentools.unpack("test")

        assert not gentools.is_iterable(async_iterable)
        assert gentools.is_iterable(iterable)

    async def test_is_iterator(self) -> None:
        async_iterator = gentools.aunpack("test").__aiter__()
        iterator = gentools.unpack("test").__iter__()

        assert not gentools.is_iterator(async_iterator)
        assert gentools.is_iterator(iterator)


class TestGenerators:
    @pytest.mark.parametrize("first,second", [(0, 1), (1, 0)])
    @pytest.mark.parametrize("order", ["little", "big"])
    @given(start=st.integers(min_value=0))
    async def test_bytes_count(
        self, start: int, order: str, first: int, second: int
    ) -> None:
        count = start
        size = (start.bit_length() // 8) + 2
        funcs = [gentools.abytes_count, gentools.bytes_count]
        async for index_0, index_1 in gentools.azip(
            funcs[first](start=start, size=size, byte_order=order),
            funcs[second](start=start, size=size, byte_order=order),
        ):
            assert index_0 == index_1
            assert index_0 == count.to_bytes(size, order)
            count += 1
            if count > start + 4:
                break

    @pytest.mark.parametrize("first,second", [(0, 1), (1, 0)])
    @given(start=st.integers())
    async def test_counts(
        self, start: int, first: int, second: int
    ) -> None:
        count = start
        funcs = [gentools.acount, gentools.count]
        async for index_0, index_1 in gentools.azip(
            funcs[first](start=start), funcs[second](start=start)
        ):
            assert index_0 == index_1
            assert index_0 == count
            count += 1
            if count > start + 4:
                break

    async def test_start_must_be_int_for_counts(self) -> None:
        problem = (  # fmt: skip
            "A non-int `start` value was allowed."
        )
        for start in (0.0, None, "test", b"test"):
            with Ignore(TypeError, if_else=violation(problem)):
                async for _ in gentools.acount(start=start):
                    pass
            with Ignore(TypeError, if_else=violation(problem)):
                for _ in gentools.count(start=start):
                    pass

    @pytest.mark.parametrize("first,second", [(0, 1), (1, 0)])
    @pytest.mark.parametrize("order", ["little", "big"])
    @pytest.mark.parametrize("size", [4, 8])
    @pytest.mark.parametrize("skip", [1, 2])
    @pytest.mark.parametrize("start", [0, 33, 256])
    async def test_bytes_ranges(
        self,
        start: int,
        skip: int,
        size: int,
        order: str,
        first: int,
        second: int,
    ) -> None:
        count = start
        end = start + 4
        funcs = [gentools.abytes_range, gentools.bytes_range]
        async for index_0, index_1 in gentools.azip(
            funcs[first](start, end, skip, size=size, byte_order=order),
            funcs[second](start, end, skip, size=size, byte_order=order),
        ):
            assert index_0 == index_1
            assert index_0 == count.to_bytes(size, order)
            assert start <= count < end
            count += skip
        assert (count - start) // skip == len([*range(start, end, skip)])

    @pytest.mark.parametrize("skip", [1, 2])
    @pytest.mark.parametrize("start", [0, 33, 256])
    async def test_ranges(self, start: int, skip: int) -> None:
        count = start
        end = start + 4
        async for index_0, index_1 in gentools.azip(
            gentools.arange(start, end, skip),
            range(start, end, skip),
        ):
            assert index_0 == index_1
            assert index_0 == count
            assert start <= count < end
            count += skip
        assert (count - start) // skip == len([*range(start, end, skip)])

    @pytest.mark.parametrize("first,second", [(0, 1), (1, 0)])
    @pytest.mark.parametrize("size", [4, 8, 13])
    @pytest.mark.parametrize("blocks", [1, 2, 3])
    async def test_batches(
        self, blocks: int, size: int, first: int, second: int
    ) -> None:
        for unit, cls in ((b"a", io.BytesIO), ("a", io.StringIO)):
            result = unit.__class__()
            data = blocks * size * unit
            funcs = [gentools.abatch, gentools.batch]
            async for block_0, block_1 in gentools.azip(
                funcs[first](data, size=size, buffer_type=cls),
                funcs[second](data, size=size, buffer_type=cls),
            ):
                assert block_0 == block_1
                assert len(block_0) == size
                result += block_0
            assert result == data

    @pytest.mark.parametrize("first,second", [(0, 1), (1, 0)])
    @pytest.mark.parametrize("unit", [[None], (None,), "a"])
    @pytest.mark.parametrize("size", [-1, 0, 1, 2, 3, 16])
    @pytest.mark.parametrize("blocks", [-1, 0, None, 1, 2, 3])
    async def test_resizes(
        self,
        blocks: t.Optional[int],
        size: int,
        unit: t.Any,
        first: int,
        second: int,
    ) -> None:
        data = 16 * unit
        result = unit.__class__()
        funcs = [gentools.aresize, gentools.resize]
        is_not_positive = lambda _: (
            (size < 1) or (isinstance(blocks, int) and blocks < 1)
        )
        with Ignore(ValueError, if_except=is_not_positive):
            item_count = 0
            async for block_0, block_1 in gentools.azip(
                funcs[first](data, size=size, blocks=blocks),
                funcs[second](data, size=size, blocks=blocks),
            ):
                item_count += (block_size := len(block_0))
                assert block_0 == block_1
                if block_size < size:
                    assert data == result + block_0
                else:
                    assert size == block_size
                result = result + block_0
                assert result == data[:item_count]
            portion = None if blocks is None else size * blocks
            assert result == data[:portion]

    @pytest.mark.parametrize("first,second", [(0, 1), (1, 0)])
    async def test_poplefts(self, first: int, second: int) -> None:
        size = 32
        count = 0
        funcs = [gentools.apopleft, gentools.popleft]
        inputs = [adeq := deque(range(size)), deq := deque(range(size))]
        async for item_0, item_1 in gentools.azip(
            funcs[first](inputs[first]), funcs[second](inputs[second])
        ):
            assert item_0 == item_1
            assert item_0 == count
            count += 1
            assert len(adeq) == len(deq)
            assert len(adeq) == size - count
        assert not adeq

    @pytest.mark.parametrize(
        "first,second,third", [(0, 1, 2), (1, 2, 0), (2, 1, 0)]
    )
    async def test_unpacks(
        self, first: int, second: int, third: int
    ) -> None:
        count = 0
        size = 32
        iterable = tuple(range(size))
        funcs = [gentools.aunpack, gentools.aunpack, gentools.unpack]
        inputs = [gentools.arange(size), iterable, iterable]
        async for index_0, index_1, index_2 in gentools.azip(
            funcs[first](inputs[first]),
            funcs[second](inputs[second]),
            funcs[third](inputs[third]),
        ):
            assert index_0 == index_1
            assert index_0 == index_2
            assert index_0 == count
            count += 1
        assert count == size

    @given(size=st.integers(min_value=0, max_value=16))
    @pytest.mark.parametrize(
        "first,second,third", [(0, 1, 2), (1, 2, 0), (2, 1, 0)]
    )
    async def test_cycles(
        self, first: int, second: int, third: int, size: int
    ) -> None:
        count = 0
        iteration = 0
        iterable = tuple(range(size))
        funcs = [gentools.acycle, gentools.acycle, gentools.cycle]
        inputs = [gentools.arange(size), iterable, iterable]
        async for index_0, index_1, index_2 in gentools.azip(
            funcs[first](inputs[first]),
            funcs[second](inputs[second]),
            funcs[third](inputs[third]),
        ):
            if size == 0:
                pytest.fail("Empty iterables produced outputs.")
            assert index_0 == index_1
            assert index_0 == index_2
            assert index_0 == count
            count = (count + 1) % size
            if count == 0:
                iteration += 1
            if iteration > 2:
                break

    @pytest.mark.parametrize("first,second", [(0, 1), (1, 0)])
    async def test_collates(self, first: int, second: int) -> None:
        arange = gentools.arange
        funcs = [gentools.acollate, gentools.collate]
        inputs = [
            (arange(0, 3), (3, 4, 5), arange(6, 9)),
            ((0, 1, 2), (3, 4, 5), (6, 7, 8)),
        ]
        count = 0
        async for index_0, index_1 in gentools.azip(
            funcs[first](*inputs[first]),
            funcs[second](*inputs[second]),
        ):
            assert index_0 == index_1
            assert index_0 == count
            count += 1
        assert count == len(inputs[1][0] + inputs[1][1] + inputs[1][2])


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import sys
from pathlib import Path


_PACKAGE_PATH = str(Path(__file__).absolute().parent.parent)
sys.path.insert(0, _PACKAGE_PATH)


from hypothesis import settings, given, strategies as st


settings.register_profile("default", max_examples=10)
settings.load_profile("default")


import json
import pytest
import hashlib
import builtins
import platform
from functools import partial
from itertools import permutations as perm, combinations as comb
from hashlib import sha3_256, sha3_512, shake_128, shake_256

import aiootp
from aiootp import *
from aiootp._paths import *
from aiootp._paths import DatabasePath
from aiootp._typing import *
from aiootp._constants.misc import *
from aiootp._constants.datasets import *
from aiootp._exceptions import *
from aiootp._gentools import aunpack, unpack, batch, abatch
from aiootp._permutations import *
from aiootp._debug_control import DebugControl
from aiootp.commons.slots import Slots, FrozenSlots
from aiootp.commons.namespaces import Namespace, OpenNamespace
from aiootp.commons.configs import Config, ConfigMap
from aiootp.asynchs import Processes, Threads, Clock
from aiootp.asynchs import run, asleep, sleep, get_process_id, get_thread_id
from aiootp.generics import *
from aiootp.generics import ByteIO
from aiootp.generics.canon import acanonical_pack, canonical_pack
from aiootp.generics.canon import acanonical_unpack, canonical_unpack
from aiootp.generics.canon import aencode_key, encode_key
from aiootp.generics.transform import axi_mix, xi_mix
from aiootp.randoms import *
from aiootp.randoms.simple import *
from aiootp.ciphers import Ciphertext, Padding

from hypothesis_strategies import identifiers
from chunky_2048_test_vectors import chunky2048_test_vector_0
from slick_256_test_vectors import slick256_test_vector_0
from passcrypt_test_vectors import (
    passcrypt_test_vector_0,
    passcrypt_test_vector_1,
    passcrypt_test_vector_2,
    passcrypt_test_vector_3,
)


# Use enable_debugging method to run tests in debug mode. May cause noticable slowdown.
DebugControl.disable_debugging()


# create static values for this test session
t = Typing


test_path = Path(__file__).parent.parent
violation = lambda problem: lambda relay: raise_exception(
    AssertionError(f"{problem} : {relay!r}")
)


class MemoizedCipher(FrozenSlots):
    __slots__ = ("config", "cipher", "salt", "aad")

    def __init__(self, cipher_type: type) -> None:
        self.config = cipher_type._config
        self.cipher = cipher_type(key=key)
        self.salt = csprng(self.config.SALT_BYTES)
        self.aad = f"testing {self.config.NAME}".encode()

    def __iter__(
        self,
    ) -> t.Generator[
        None, t.Union[t.ConfigType, t.CipherInterfaceType, bytes], None
    ]:
        yield self.config
        yield self.cipher
        yield self.salt
        yield self.aad


key = csprng(168)
dual_output_ciphers = [MemoizedCipher(Chunky2048)]
shake_permute_ciphers = [MemoizedCipher(Slick256)]
all_ciphers = [*dual_output_ciphers, *shake_permute_ciphers]
dual_output_cipher_names = [conf.NAME for (conf, *_) in dual_output_ciphers]
shake_permute_cipher_names = [
    conf.NAME for (conf, *_) in shake_permute_ciphers
]


passphrase_0 = csprng()
passphrase_1 = randoms.token_bytes(32)
passphrases = [passphrase_0, passphrase_1]


salt_0 = csprng()
salt_1 = randoms.token_bytes(32)
salts = [salt_0, salt_1]


passcrypt_settings = OpenNamespace(mb=1, cpu=1, cores=1)


tag = "testing"
atag = "a" + tag
metatag = "clients"
ametatag = "a" + metatag


username = b"test suite"
passphrase = b"terrible low entropy passphrase"
PROFILE = dict(
    username=username,
    passphrase=passphrase,
    salt=salt_0,
    aad=b"testing passcrypt",
)
LOW_PASSCRYPT_SETTINGS = OpenNamespace(mb=1, cpu=1, cores=1)
PROFILE_AND_SETTINGS = {**PROFILE, **LOW_PASSCRYPT_SETTINGS}


byte_leakage = 16 * b"\x00"
string_leakage = 16 * b"0"
plaintext_bytes = b"\xff" + 256 * b"\x00"
plaintext_string = "f" + 256 * "0"
test_data = {
    "floats": 10000.243,
    "ints": list(range(-16, 16, 8)),
    "dicts": {"testing": {}},
    "lists": 4 * [[]],
    "strings": "testing...",
    "nul": None,
    "bools": [True, False],
}
atest_data = test_data.copy()


# Creating timestamped values early so later testing has to wait less time
ttl_test_cipher = choice(all_ciphers)
atest_json_ciphertext = run(
    ttl_test_cipher.cipher.ajson_encrypt(
        atest_data, aad=ttl_test_cipher.aad
    )
)
test_json_ciphertext = ttl_test_cipher.cipher.json_encrypt(
    test_data, aad=ttl_test_cipher.aad
)
atest_token_ciphertext = run(
    ttl_test_cipher.cipher.amake_token(
        plaintext_bytes, aad=ttl_test_cipher.aad
    )
)
test_token_ciphertext = ttl_test_cipher.cipher.make_token(
    plaintext_bytes, aad=ttl_test_cipher.aad
)


async def make_async_ttl_cipher_stream() -> t.Tuple[
    t.AsyncCipherStream, bytes
]:
    stream = await ttl_test_cipher.cipher.astream_encrypt(
        salt=ttl_test_cipher.salt, aad=ttl_test_cipher.aad
    )
    await stream.abuffer(plaintext_bytes)
    ciphertext = b"".join(
        [(block_id + block) async for block_id, block in stream.afinalize()]
    )
    return stream, ciphertext


attl_cipher_stream, attl_stream_ciphertext = run(
    make_async_ttl_cipher_stream()
)

ttl_cipher_stream = ttl_test_cipher.cipher.stream_encrypt(
    salt=ttl_test_cipher.salt, aad=ttl_test_cipher.aad
)
ttl_cipher_stream.buffer(plaintext_bytes)
ttl_stream_ciphertext = b"".join(
    (block_id + block) for block_id, block in ttl_cipher_stream.finalize()
)


light_pcrypt = Passcrypt(mb=1, cpu=1, cores=1, tag_size=32)
aexpired_passcrypt_hash = run(light_pcrypt.ahash_passphrase(passphrase_0))
expired_passcrypt_hash = light_pcrypt.hash_passphrase(passphrase_0)


clock = Clock(SECONDS, epoch=EPOCH_NS)
ns_clock = Clock(NANOSECONDS, epoch=EPOCH_NS)
time_start = clock.make_timestamp()


@pytest.fixture(scope="session")
def database() -> Database:
    db = Database(key=key, preload=True)
    db.save_database()
    yield db

    db.delete_database()


@pytest.fixture(scope="session")
def async_database() -> AsyncDatabase:
    db = run(AsyncDatabase(key=key, preload=True))
    yield db

    run(db.adelete_database())


@pytest.fixture
def path() -> Path:
    file_path = Path("byte_io_testing_path.txt").absolute()

    if not file_path.is_file():
        file_path.write_bytes(b"")

    yield file_path

    if file_path.is_file():
        file_path.unlink()


@pytest.fixture
def salt_path() -> Path:
    parent_dir = DatabasePath() / "secure"
    file_path = parent_dir / "salt_testing_path.txt"

    yield file_path

    if file_path.is_file():
        file_path.chmod(0o600)
        file_path.unlink()


class ExampleConfig(Config):
    __slots__ = ("NUMBER", "STRING")

    slots_types = dict(NUMBER=int, STRING=str)

    def __init__(self, number: int, string: str) -> None:
        self.NUMBER = number
        self.STRING = string


@pytest.fixture
def config() -> ExampleConfig:
    return ExampleConfig(number=420, string="word")


@pytest.fixture
def mapping() -> ConfigMap:
    return ConfigMap(config_type=ExampleConfig)


@pytest.fixture(scope="session")
def pkg_context() -> Namespace:
    context = OpenNamespace(
        test_path=test_path,
        signing_key=PackageSigner.generate_signing_key(),
    )
    context.signer_init = OpenNamespace(
        package=aiootp.__package__,
        version=aiootp.__version__,
        author=aiootp.__author__,
        license=aiootp.__license__,
        description=aiootp.__doc__,
        date=Clock(DAYS, epoch=0).time(),
        build_number=0,
    )
    context.signer_db_init = OpenNamespace(
        username=b"test_username",
        salt=token_bytes(64),
        passphrase=token_bytes(64),
        path=DatabasePath(),
        **LOW_PASSCRYPT_SETTINGS,
    )
    context.update(**context.signer_init, **context.signer_db_init)
    filename_sheet = """
    include tests/conftest.py
    include tests/test_ByteIO.py
    include tests/test_Database_AsyncDatabase.py
    include tests/test_Passcrypt.py
    include tests/test_StreamHMAC.py
    include tests/test_X25519_Ed25519.py
    include tests/all_aiootp_tests.py
    include tests/all_ciphers_tests.py
    include tests/all_generics_tests.py
    include tests/test_high_level_encryption.py
    include tests/test_misc_in_ciphers.py
    include tests/test_misc_in_generics.py
    include tests/test_misc_in_randoms.py
    include tests/all_randoms_tests.py
    """.strip().split("\n")
    context.files = [
        test_path / line.strip().split(" ")[-1] for line in filename_sheet
    ]
    return context


@pytest.fixture(scope="session")
def pkg_signer(pkg_context: Namespace) -> PackageSigner:
    signer = PackageSigner(**pkg_context.signer_init)
    is_mac_os_issue = lambda _: (platform.system() == "Darwin")
    while True:
        sleep(0.001)
        with Ignore(ConnectionRefusedError, if_except=is_mac_os_issue):
            signer.connect_to_secure_database(**pkg_context.signer_db_init)
            break

    signer.update_signing_key(pkg_context.signing_key)
    for path in pkg_context.files:
        with path.open("rb") as source_file:
            signer.add_file(str(path), source_file.read())

    signer.sign_package()
    yield signer
    signer.db.delete_database()


@pytest.fixture(scope="session")
def pkg_verifier(pkg_signer: PackageSigner) -> PackageVerifier:
    public_bytes = pkg_signer.signing_key.public_bytes
    return PackageVerifier(public_bytes, path=test_path)


__all__ = [n for n in globals() if not n.startswith("_")]



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import os
import warnings

from aiootp.asynchs import aos

from conftest import *


OS_VARNAMES: t.FrozenSet[str] = frozenset(dir(os))


def name_will_proc_warning(name: str) -> bool:
    return name not in OS_VARNAMES


@given(name=identifiers().filter(name_will_proc_warning))
async def test_not_implemented_placeholder(name: str) -> None:
    with warnings.catch_warnings(record=True) as warning:
        await aos._not_implemented_placeholder(name)()

        assert isinstance(warning[-1].message, UserWarning)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A multi-threading interface.
"""

__all__ = ["Threads", "get_thread_id"]


import queue
from threading import Thread
from concurrent.futures import thread
from _thread import get_ident as get_thread_id
from concurrent.futures import ThreadPoolExecutor

from aiootp._typing import Typing as t

from .concurrency_interface import ConcurrencyInterface


class Threads(ConcurrencyInterface):
    """
    Simplifies spawning & returning the values produced by `Thread` &
    `ThreadPoolExecutor` objects with an interface over these types from
    the `threading` & `concurrent.futures` packages.
    """

    __slots__ = ()

    _default_probe_delay: float = 0.001
    _pool: t.PoolExecutorType = ThreadPoolExecutor()
    _type: type = Thread

    BrokenPool: type = thread.BrokenThreadPool

    get_id: t.Callable[[], int] = get_thread_id

    @classmethod
    def _get_queue(cls, /, maxsize: int = 1) -> t.QueueType:
        """
        Returns a queue object to retrieve values from spawned workers.
        """
        return queue.Queue(maxsize=maxsize)

    @classmethod
    def reset_pool(cls, /) -> None:
        """
        When a thread pool is broken by being abruptly exited, this
        method can be called to reset the class' pool object with a new
        instance.
        """
        cls._pool = cls._pool.__class__()


module_api = dict(
    Threads=t.add_type(Threads),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    get_thread_id=get_thread_id,
)



#!/usr/bin/env python3
#
# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A script to run a service that signs build data sent via local socket
requests.

 _____________________________________
|                                     |
|            Usage Example:           |
|_____________________________________|

Terminal #1:

~$ source <path/to/secrets/env/file.sh>
~$ uv run sign_package_service.py  # CLI that awaits requests

Terminal #2:

~$ source <path/to/secrets/env/file.sh>
~$ uv run sign_package_build.py  # CLI to request a new signature
"""

import io
import os
import json
import socket
from getpass import getpass
from contextlib import contextmanager

from aiootp import __package__, __version__
from aiootp import Chunky2048 as Cipher
from aiootp import PackageSigner
from aiootp._typing import Typing as t
from aiootp.generics import canonical_pack, canonical_unpack


HOST: str = "localhost"
PORT: int = 13120
TTL: float = 2.0
EXIT_TTL: float = 7200.0
MAX_BUFFER: int = 8192
RECV_AAD: bytes = b"build_signature_request"
SEND_AAD: bytes = b"provided_build_signature"
HEADER_SLICE: slice = Cipher._config.HEADER_SLICE


IDLE_SHUTDOWN_SIGNALS: t.Tuple[type] = (
    KeyboardInterrupt,
    SystemExit,
    TimeoutError,
)
EXIT_SIGNALS: t.Tuple[type] = (
    KeyboardInterrupt,
    SystemExit,
)
CHANNEL_ERRORS: t.Tuple[type] = (
    ConnectionAbortedError,
    ConnectionResetError,
    TimeoutError,
)
PROCESSING_ERRORS: t.Tuple[type] = (
    TypeError,
    ValueError,
    json.JSONDecodeError,
)
INTEGRITY_FAILURES: t.Tuple[type] = (
    Cipher.InvalidSHMAC,
    Cipher.TimestampExpired,
)


_KEY_NEGOTIATION_FEATURE_FLAG: bool = False


def current_transmit_key(
    channel: t.Optional[socket.socket] = None, context: t.Any = None
) -> bytes:
    """
    Supports future key rotation implementations by querying for the
    transmission key at the start of each signature request.

    Returns the current bytes-type key.

    **Under Development**: Key negotiation may be implemented in future.
    The parameters signal this possibility, but are currently unused.
    """
    if _KEY_NEGOTIATION_FEATURE_FLAG and (channel or context):
        raise NotImplementedError("Key negotiation not available.")
    return bytes.fromhex(os.getenv("_TRANSMIT_KEY"))


def update_signing_key_prompt(signer: PackageSigner) -> None:
    """
    Provides a CLI flow to optionally insert the hex package signing key
    into the package signer's database.
    """
    if (
        getpass("is the signing key already saved on this device? (Y/n) ")
        .lower()
        .strip()
        .startswith("n")
    ):
        signer.update_signing_key(
            bytes.fromhex(getpass("signing key: ").strip())
        )


def update_public_credentials_prompt(signer: PackageSigner) -> None:
    """
    Provides a CLI flow to optionally insert the {str_name: str_value}
    public credentials into the package signer's database.
    """
    while (
        getpass("update public credentials? (y/N) ")
        .lower()
        .strip()
        .startswith("y")
    ):
        signer.update_public_credentials(
            **{getpass("name: ").strip(): getpass("value: ")}
        )


def make_signer_object() -> PackageSigner:
    """
    Provides a CLI flow to unlock the package signer's database, &
    optionally insert missing or outdated values to be included in
    signature contexts.

    Returns the package signer object.
    """
    signer = PackageSigner(package=__package__, version=__version__)
    signer.connect_to_secure_database(
        username=getpass("database username: ").encode(),
        passphrase=getpass("database key: ").encode(),
        salt=getpass("database salt: ").encode(),
        path=getpass("secure directory: "),
    )
    update_signing_key_prompt(signer)
    update_public_credentials_prompt(signer)
    return signer


@contextmanager
def start_server(
    host: str = HOST, port: int = PORT, *, max_channels: int = 1
) -> socket.socket:
    """
    Wraps a new server socket listening on `host`:`port`, in a context
    guaranteed to close the socket when the context ends.

    Yields the new server socket.
    """
    try:
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.settimeout(EXIT_TTL)
        server.bind((host, port))
        server.listen(max_channels)
        yield server
    finally:
        server.close()


def buffer_recv(channel: socket.socket, size: int) -> bytes:
    """
    A reliable method of ingesting exactly `size` number of bytes from a
    socket `channel`.

    Returns the received bytes.
    """
    index = 0
    buffer = io.BytesIO()
    while index < size:
        delta = buffer.write(channel.recv(min(size - index, MAX_BUFFER)))
        if not delta:
            raise ConnectionAbortedError("The channel has been closed.")
        index += delta
        buffer.seek(index)
    buffer.seek(0)
    return buffer.read()


def buffer_send(channel: socket.socket, data: bytes) -> int:
    """
    A reliable method of sending all of the `data` bytes through a
    socket `channel` while still enforcing a max buffer size.

    Returns the int number of bytes which were sent.
    """
    size = len(data)
    channel.sendall(size.to_bytes(8, "big"))
    buffer = io.BytesIO(data)
    while chunk := buffer.read(MAX_BUFFER):
        channel.sendall(chunk)
    return size


def get_and_parse_request(
    channel: socket.socket, cipher: t.CipherInterfaceType
) -> t.Tuple[bytes, t.Dict[str, t.JSONSerializable], t.Dict[str, str]]:
    """
    Authenticates the received request data from a `channel` using a
    transmission key shared between the service & the requester.

    Returns the relevant parsed parts of the request.
    """
    recv_length = int.from_bytes(buffer_recv(channel, 8), "big")
    request = buffer_recv(channel, recv_length)
    signing_request = cipher.bytes_decrypt(request, aad=RECV_AAD, ttl=TTL)
    scope, files = canonical_unpack(signing_request)
    return request[HEADER_SLICE], json.loads(scope), json.loads(files)


def produce_signed_summary(
    signer: t.PackageSigner,
    scope: t.Dict[str, t.JSONSerializable],
    files: t.Dict[str, str],
) -> bytes:
    """
    Ingests the package `scope` & `files` metadata into the `signer`.

    Returns the JSON serialized signed bytes-type summary.
    """
    signer.__init__(**scope)
    signer.files.update(files)
    signer.sign_package()
    return json.dumps(signer.summarize(), indent=4).encode()


def send_packaged_response(
    channel: socket.socket,
    cipher: t.CipherInterfaceType,
    header: bytes,
    summary: bytes,
) -> int:
    """
    Sends the requested context data & signature `summary` through the
    `channel` to the requester, within a transcript-specific ciphertext.

    Returns the number of bytes which were sent.
    """
    aad = canonical_pack(SEND_AAD, header)
    response = cipher.bytes_encrypt(summary, aad=aad)
    return buffer_send(channel, response)


def signing_service_loop(
    signer: PackageSigner,
    server: socket.socket,
    get_transmit_key: t.Callable[..., bytes],
) -> None:
    """
    Awaits & fulfills authentic signature requests sent over a socket.
    """
    while True:
        try:
            channel, _ = server.accept()
        except IDLE_SHUTDOWN_SIGNALS:
            break
        try:
            channel.settimeout(TTL)
            cipher = Cipher(get_transmit_key())

            header, scope, files = get_and_parse_request(channel, cipher)
            if scope["package"] != __package__:
                raise ValueError("Invalid context switch.")

            summary = produce_signed_summary(signer, scope, files)
            send_packaged_response(channel, cipher, header, summary)
        except EXIT_SIGNALS:
            break
        except CHANNEL_ERRORS as e:
            print(f"Connection issue encountered: {e!r}")
        except PROCESSING_ERRORS as e:
            print(f"Message issue encountered: {e!r}")
        except INTEGRITY_FAILURES as e:
            print(f"Invalid request received: {e!r}")
        finally:
            channel.close()


def run_signing_service(
    host: str = HOST,
    port: int = PORT,
    *,
    get_transmit_key: t.Callable[..., bytes],
) -> None:
    """
    Binds the signing service to a socket at the `host` and `port`,
    where it will receive signature requests & authenticate them with
    the symmetric key returned from the `get_transmit_key` callable
    argument.
    """
    signer = make_signer_object()
    with start_server(host=host, port=port) as server:
        signing_service_loop(signer, server, get_transmit_key)


if __name__ == "__main__":
    run_signing_service(get_transmit_key=current_transmit_key)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


class TestPackageVerifier:
    async def test_path_needed_to_verify_file_hashes(
        self, pkg_signer: PackageSigner
    ) -> None:
        problem = (  # fmt: skip
            "The default `verify_files=True` was allowed to initialize "
            "without specifying a relative directory to locate files."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            PackageVerifier(pkg_signer.signing_key.public_bytes)

        verifier = PackageVerifier(
            pkg_signer.signing_key.public_bytes, verify_files=False
        )
        verifier.verify_summary(pkg_signer.summarize())

    async def test_if_path_given_verify_files_must_be_true(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        problem = (  # fmt: skip
            "A verifier was initialized with conflicting flags."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            PackageVerifier(
                pkg_signer.signing_key.public_bytes,
                path=pkg_context.test_path,
                verify_files=False,
            )

    async def test_verifier_detects_wrong_file_digest(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        verifier = PackageVerifier(
            pkg_signer.signing_key.public_bytes, path=pkg_context.test_path
        )
        summary = pkg_signer.summarize()
        filename = randoms.choice(list(summary[CHECKSUMS]))
        fake_hashed_file = pkg_signer._Hasher(csprng())
        summary[CHECKSUMS][filename] = fake_hashed_file.hexdigest()
        problem = (  # fmt: skip
            f"An invalid digest for file {filename=} wasn't detected."
        )
        with Ignore(InvalidDigest, if_else=violation(problem)):
            verifier._verify_file_checksums(summary)

    async def test_signing_key_input_acceptable_types(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        summary = pkg_signer.summarize()
        public_keys = [
            Ed25519().import_public_key(pkg_signer.signing_key.public_key),
            pkg_signer.signing_key.public_key,
            pkg_signer.signing_key.public_bytes,
        ]
        for key in public_keys:
            verifier = PackageVerifier(key, path=pkg_context.test_path)
            verifier.verify_summary(summary)

    async def test_altering_signing_key_fails(
        self,
        pkg_context: Namespace,
        pkg_signer: PackageSigner,
        pkg_verifier: PackageVerifier,
    ) -> None:
        summary = pkg_signer.summarize()
        wrong_summary = summary.copy()
        wrong_key = X25519().generate().public_bytes
        wrong_summary["signing_key"] = wrong_key.hex()
        wrong_verifier = PackageVerifier(
            wrong_key, path=pkg_context.test_path
        )

        problem = (  # fmt: skip
            "An altered `signing_key` went uncaught."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            wrong_verifier.verify_summary(summary)
        with Ignore(ValueError, if_else=violation(problem)):
            wrong_verifier.verify_summary(wrong_summary)
        with Ignore(ValueError, if_else=violation(problem)):
            pkg_verifier.verify_summary(wrong_summary)

    async def test_altering_checksum_fails(
        self,
        pkg_signer: PackageSigner,
        pkg_verifier: PackageVerifier,
    ) -> None:
        summary = pkg_signer.summarize()
        summary["checksum"] = summary["checksum"][::-1]
        problem = (  # fmt: skip
            "Summary alteration uncaught."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            pkg_verifier.verify_summary(summary)

        # returning the checksum to the original value succeeds
        summary["checksum"] = summary["checksum"][::-1]
        pkg_verifier.verify_summary(summary)

        # a json summary also works
        pkg_verifier.verify_summary(json.dumps(summary))


class TestPackageSigner:
    async def test_changing_signature_detected_during_summarization(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        PACKAGE = pkg_context.package
        VERSIONS = pkg_signer._VERSIONS
        VERSION = pkg_context.version
        signature = pkg_signer.db[PACKAGE][VERSIONS][VERSION]
        pkg_signer.db[PACKAGE][VERSIONS][VERSION] = token_bytes(64).hex()

        problem = (  # fmt: skip
            "Altered signature not detected during summarization."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            pkg_signer.summarize()

        # returning the signature works
        pkg_signer.db[PACKAGE][VERSIONS][VERSION] = signature
        pkg_signer.summarize()

    async def test_signing_key_in_summary_is_hex_public_key(
        self, pkg_signer: PackageSigner
    ) -> None:
        summary = pkg_signer.summarize()
        assert len(summary["signing_key"]) == 64
        assert (
            summary["signing_key"]
            == pkg_signer.signing_key.public_bytes.hex()
        )

    async def test_altering_the_signature_summary_fails(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        summary = pkg_signer.summarize()
        verifier = PackageVerifier(
            pkg_signer.signing_key.public_bytes, path=pkg_context.test_path
        )

        problem = (  # fmt: skip
            "Signature alteration went uncaught."
        )
        with Ignore(verifier.InvalidSignature, if_else=violation(problem)):
            fake_summary = {**summary, "signature": token_bytes(64).hex()}
            verifier.verify_summary(fake_summary)

    async def test_unmodified_signer_summary_is_deterministic(
        self, pkg_signer: PackageSigner
    ) -> None:
        summary = pkg_signer.summarize()
        for _ in range(2):
            assert summary == pkg_signer.summarize()
            assert json.dumps(summary) == json.dumps(pkg_signer.summarize())
            assert summary == json.loads(json.dumps(pkg_signer.summarize()))

    async def test_scope_items_show_in_repr(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        _repr = repr(pkg_signer)
        assert all(
            f"{name}={pkg_context[name]!r}" in _repr
            for name in pkg_signer._scope
        )

    @given(
        build_number=st.one_of(
            st.none(),
            st.booleans(),
            st.integers(),
            st.floats(),  # if nan, can't directly check equality
            st.text(),
        ).filter(lambda x: x != 0)
    )
    async def test_scope_updates_when_instructed(
        self,
        pkg_context: Namespace,
        pkg_signer: PackageSigner,
        build_number: t.Hashable,
    ) -> None:
        old_number = pkg_context.build_number
        assert 2 == len({pkg_signer._scope.build_number, build_number})
        assert 1 == len({pkg_signer._scope.build_number, old_number})
        pkg_signer.update_scope(build_number=build_number)
        assert 1 == len({pkg_signer._scope.build_number, build_number})
        pkg_signer.sign_package()

        pkg_signer.update_scope(build_number=old_number)
        assert 1 == len({pkg_signer._scope.build_number, old_number})
        pkg_signer.sign_package()

    async def test_file_hashes_are_stored_after_being_added(
        self,
        pkg_context: Namespace,
        pkg_signer: PackageSigner,
        pkg_verifier: PackageVerifier,
    ) -> None:
        pkg_signer.sign_package()
        pkg_verifier.verify_summary(summary := pkg_signer.summarize())
        signer = PackageSigner(**pkg_context.signer_init)
        signer.connect_to_secure_database(**pkg_context.signer_db_init)
        signer.update_signing_key(pkg_signer.signing_key)
        assert signer.db._root_filename == pkg_signer.db._root_filename
        for index, path in enumerate(pkg_context.files, start=1):
            filename = str(path)
            file_data = path.read_bytes()
            signer.add_file(filename, file_data)
            assert index == len(signer.files)
            assert filename in signer.files
            assert signer.files[filename] != signer._Hasher().hexdigest()
            assert (
                signer.files[filename]
                == signer._Hasher(file_data).hexdigest()
            )
        assert len(signer.files) == len(pkg_context.files)
        signer.sign_package()
        pkg_verifier.verify_summary(new_summary := signer.summarize())
        assert summary == new_summary

    async def test_cant_retrieve_signing_key_prior_to_setting(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        signing_key = pkg_signer.signing_key
        pkg_signer.db[pkg_context.package]["versions"] = {}
        try:
            problem = (  # fmt: skip
                "Allowed to retrieve a signing key before its creation."
            )
            del pkg_signer.db[pkg_context.package]["signing_key"]
            assert not hasattr(pkg_signer, "signing_key")
            pkg_signer.db[pkg_context.package]["signing_key"] = ""
            assert not hasattr(pkg_signer, "signing_key")
            with Ignore(t.SigningKeyNotSet, if_else=violation(problem)):
                assert pkg_signer.signing_key
        finally:
            pkg_signer.update_signing_key(signing_key)
            pkg_signer.sign_package()

    async def test_update_signing_key_interface_changes_signing_key(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        summary = pkg_signer.summarize()
        signing_key = pkg_signer.signing_key
        try:
            new_signing_key = pkg_signer.generate_signing_key()
            pkg_signer.update_signing_key(new_signing_key)
            pkg_signer.sign_package()
            new_summary = pkg_signer.summarize()
            verifier = PackageVerifier(
                new_signing_key, path=pkg_context.test_path
            )
            assert summary["signing_key"] != new_summary["signing_key"]
            assert (
                signing_key.public_bytes
                != pkg_signer.signing_key.public_bytes
            )
            verifier.verify_summary(new_summary)
        finally:
            pkg_signer.update_signing_key(signing_key)
            pkg_signer.sign_package()

    async def test_update_signing_key_interface_acceptable_types(
        self, pkg_signer: PackageSigner
    ) -> None:
        # update_signing_key expects a bytes, Ed25519PrivateKey or an Ed25519 object
        pkg_signer.update_signing_key(pkg_signer.signing_key)
        pkg_signer.update_signing_key(pkg_signer.signing_key.secret_key)
        pkg_signer.update_signing_key(pkg_signer.signing_key.secret_bytes)

        problem = (  # fmt: skip
            "A type other than bytes, Ed25519PrivateKey or an "
            "Ed25519 object was allowed for update_signing_key."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            pkg_signer.update_signing_key(
                bytearray(pkg_signer.signing_key.secret_bytes)
            )
        with Ignore(TypeError, if_else=violation(problem)):
            pkg_signer.update_signing_key(
                pkg_signer.signing_key.secret_bytes.hex()
            )

    async def test_cant_retrieve_signature_prior_to_signing(
        self, pkg_context: Namespace, pkg_signer: PackageSigner
    ) -> None:
        del pkg_signer.db[pkg_context.package]["versions"][
            pkg_context.version
        ]
        problem = (  # fmt: skip
            "Allowed to retrieve a signature before signing."
        )
        assert not hasattr(pkg_signer, "_signature")
        with Ignore(t.PackageNotSigned, if_else=violation(problem)):
            assert pkg_signer._signature
        pkg_signer.sign_package()
        assert pkg_signer._signature

    async def test_update_public_credentials_updates_state(
        self, pkg_signer: PackageSigner
    ) -> None:
        public_credentials = pkg_signer.db["aiootp"]["public_credentials"]
        summary = pkg_signer.summarize()
        assert not public_credentials
        assert "x25519_public_key" not in public_credentials
        pkg_signer.update_public_credentials(
            x25519_public_key=aiootp.__PUBLIC_X25519_KEY__
        )
        pkg_signer.sign_package()
        new_summary = pkg_signer.summarize()
        assert summary != new_summary
        assert public_credentials
        assert (
            aiootp.__PUBLIC_X25519_KEY__
            == public_credentials["x25519_public_key"]
        )

    def test_database_must_be_connected_to_retrieve_signer_state(
        self, pkg_context: Namespace
    ) -> None:
        signer = PackageSigner(**pkg_context.signer_init)
        problem = (  # fmt: skip
            "Allowed to retrieve database states before connecting to the "
            "database."
        )
        assert not hasattr(signer, "_db")
        with Ignore(t.DatabaseNotConnected, if_else=violation(problem)):
            assert signer.db


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


INTERFACES = (Processes, Threads)


@pytest.mark.parametrize("interface", INTERFACES)
async def test_anew_workers_have_distinct_ids_from_their_caller(
    interface: t.ConcurrencyInterface,
) -> None:
    amethod = interface.aget_id
    method = interface.get_id

    result = await interface.anew(amethod)
    acaller_id = await amethod()
    caller_id = method()
    assert result > 0
    assert result.__class__ is int
    assert result != acaller_id
    assert result != caller_id
    assert acaller_id == caller_id

    result = await interface.anew(method)
    acaller_id = await amethod()
    caller_id = method()
    assert result > 0
    assert result.__class__ is int
    assert result != acaller_id
    assert result != caller_id
    assert acaller_id == caller_id


@pytest.mark.parametrize("interface", INTERFACES)
async def test_new_workers_have_distinct_ids_from_their_caller(
    interface: t.ConcurrencyInterface,
) -> None:
    amethod = interface.aget_id
    method = interface.get_id

    result = interface.new(method)
    acaller_id = await amethod()
    caller_id = method()
    assert result > 0
    assert result.__class__ is int
    assert result != acaller_id
    assert result != caller_id
    assert acaller_id == caller_id


@pytest.mark.parametrize("interface", INTERFACES)
async def test_asubmit_workers_have_distinct_ids_from_their_caller(
    interface: t.ConcurrencyInterface,
) -> None:
    interface.reset_pool()
    amethod = interface.aget_id
    method = interface.get_id

    fut = await interface.asubmit(amethod)
    result = await fut.aresult()
    acaller_id = await amethod()
    caller_id = method()
    assert result == fut.result()
    assert result > 0
    assert result.__class__ is int
    assert result != acaller_id
    assert result != caller_id
    assert acaller_id == caller_id

    fut = await interface.asubmit(method)
    result = await fut.aresult()
    acaller_id = await amethod()
    caller_id = method()
    assert result == fut.result()
    assert result > 0
    assert result.__class__ is int
    assert result != acaller_id
    assert result != caller_id
    assert acaller_id == caller_id


@pytest.mark.parametrize("interface", INTERFACES)
async def test_submit_workers_have_distinct_ids_from_their_caller(
    interface: t.ConcurrencyInterface,
) -> None:
    interface.reset_pool()
    amethod = interface.aget_id
    method = interface.get_id

    fut = interface.submit(method)
    result = fut.result()
    acaller_id = await amethod()
    caller_id = method()
    assert result > 0
    assert result.__class__ is int
    assert result != acaller_id
    assert result != caller_id
    assert acaller_id == caller_id


@pytest.mark.parametrize("interface", INTERFACES)
async def test_probe_delay_must_be_positive(
    interface: t.ConcurrencyInterface,
) -> None:
    amethod = interface.aget_id
    method = interface.get_id

    problem = (  # fmt: skip
        "A non-positive probe_delay was allowed."
    )
    with Ignore(ValueError, if_else=violation(problem)):
        await interface.anew(amethod, probe_delay=-1)

    with Ignore(ValueError, if_else=violation(problem)):
        interface.new(method, probe_delay=-1)

    with Ignore(ValueError, if_else=violation(problem)):
        await interface.asubmit(amethod, probe_delay=-1)

    with Ignore(ValueError, if_else=violation(problem)):
        interface.submit(method, probe_delay=-1)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



#!/usr/bin/env python3
#
# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A script to package build data for verification, or for a request of
signature from the package signer service via a local socket.

 _____________________________________
|                                     |
|            Usage Example:           |
|_____________________________________|

Terminal #1:

~$ source <path/to/secrets/env/file.sh>
~$ uv run sign_package_service.py  # CLI that awaits requests

Terminal #2:

~$ source <path/to/secrets/env/file.sh>
~$ uv run sign_package_build.py  # CLI to request a new signature
"""

import json
import socket
from shlex import split
from pathlib import Path
from getpass import getpass
from subprocess import check_output
from contextlib import contextmanager

from aiootp import (
    __doc__,
    __license__,
    __package__,
    __version__,
    __author__,
    __PUBLIC_ED25519_KEY__,
)
from aiootp import PackageSigner, PackageVerifier
from aiootp._typing import Typing as t
from aiootp.generics import canonical_pack

from sign_package_service import PORT, TTL, HEADER_SLICE
from sign_package_service import buffer_send, buffer_recv
from sign_package_service import Cipher, current_transmit_key


HOST: str = "localhost"
SEND_AAD: bytes = b"build_signature_request"
RECV_AAD: bytes = b"provided_build_signature"
SIGNING_KEY: bytes = bytes.fromhex(__PUBLIC_ED25519_KEY__)


MANIFEST_EXCLUDES: t.Set[str] = {
    ".github/",
    "aiootp/db/",
    "aiootp/tor/",
}
DYNAMIC_FILES: t.Set[str] = {
    (MANIFEST_FILENAME := "MANIFEST.in"),
    (SIGNATURE_FILENAME := "SIGNATURE.txt"),
}


Hasher: t.HasherType = PackageSigner._Hasher


def discover_git_branch_tree() -> t.Tuple[str, str]:
    """
    Calls & parses git responses to determine the current branch or ref
    name, & an enumeration of all the filenames tracked in the repo.

    Returns the str git branch & str list of branch & staged filenames.
    """
    branch_command = split("git rev-parse --abbrev-ref HEAD")
    git_branch = check_output(branch_command).strip()
    if b"* " + git_branch not in check_output(split("git branch")):
        print(f"\nNOTICE: {git_branch=} isn't an existing branch.")
        git_branch = getpass("\ndesired branch: ").encode() or git_branch

    branch_tree_command = split("git ls-files --cached")
    git_branch_tree = check_output(branch_tree_command).strip()
    return git_branch.decode(), git_branch_tree.decode()


def update_scope_prompt(branch: str) -> t.Dict[str, t.JSONSerializable]:
    """
    Gathers package metadata to be added into the signature context, &
    offers the option via CLI to update the build number.

    Returns the metadata as a JSON object-like dict.
    """
    print(f"\ncurrent version: {__version__}")

    if (old_signature_file := Path(SIGNATURE_FILENAME)).is_file():
        old_signature = json.loads(old_signature_file.read_bytes())
        old_scope = old_signature[PackageSigner._SCOPE]
        old_build_number = old_scope["build_number"]
        print(f"current build: {old_build_number}\n")

    return dict(
        package=__package__,
        version=__version__,
        author=__author__,
        license=__license__,
        description=__doc__,
        git_branch=branch,
        build_number=getpass("build number: ") or old_build_number,
    )


def discover_file_inventory(
    git_branch_tree: str,
) -> t.Tuple[bytes, t.Dict[str, str]]:
    """
    Dynamically reconstructs the Python 'MANIFEST.in' from the git file
    list, excluding external data directories. The appropriate filenames
    are bundled along with their source data's hexdigests into a dict to
    be sent over to the signing service.

    Returns the new manifest file, and the bundled file metadata.
    """
    manifest = ""
    files = {}

    for line in git_branch_tree.split("\n"):
        if not (filename := line.strip()):
            continue
        if all((part not in filename) for part in MANIFEST_EXCLUDES):
            manifest += f"include {filename}\n"
        if filename not in DYNAMIC_FILES:
            hashed_file = Hasher(Path(filename).read_bytes())
            files[filename] = hashed_file.hexdigest()

    manifest = manifest.encode()
    files[MANIFEST_FILENAME] = Hasher(manifest).hexdigest()
    return manifest, files


@contextmanager
def start_client(host: str = HOST, port: int = PORT) -> socket.socket:
    """
    Wraps a new client socket at `host`:`port` in a context guaranteed
    to close the channel when the context ends.

    Yields the new client socket.
    """
    try:
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        client.settimeout(TTL)
        client.connect((host, port))
        yield client
    finally:
        client.close()


def send_request(
    client: socket.socket,
    cipher: t.CipherInterfaceType,
    scope: t.Dict[str, t.JSONSerializable],
    files: t.Dict[str, str],
) -> bytes:
    """
    Wraps the package `scope` & `files` metadata within authenticated
    encryption from the `cipher`, & sends the ciphertext request over
    the `client` socket to the service.

    Returns the request ciphertext header, which is used later to bind
    the service response ciphertext to the current request.
    """
    signing_request = canonical_pack(
        json.dumps(scope).encode(), json.dumps(files).encode()
    )
    request = cipher.bytes_encrypt(signing_request, aad=SEND_AAD)
    buffer_send(client, request)
    return request[HEADER_SLICE]


def recv_response(
    client: socket.socket, cipher: t.CipherInterfaceType, header: bytes
) -> bytes:
    """
    Receives a service response from the `client` socket, then decrypts
    & authenticates the ciphertext with the `cipher`.

    Returns the resulting signed package summary.
    """
    recv_length = int.from_bytes(buffer_recv(client, 8), "big")
    response = buffer_recv(client, recv_length)
    aad = canonical_pack(RECV_AAD, header)
    return cipher.bytes_decrypt(response, aad=aad, ttl=TTL)


def request_package_summary_signature(
    get_transmit_key: t.Callable[..., bytes],
) -> bytes:
    """
    Bundles the package metadata & requests a signature from the signing
    service over a socket connection. The channel's integrity is ensured
    by the key returned from the `get_transmit_key` callable argument.

    Returns the resulting signed package summary.
    """
    git_branch, git_branch_tree = discover_git_branch_tree()

    scope = update_scope_prompt(git_branch)
    manifest, files = discover_file_inventory(git_branch_tree)

    Path(MANIFEST_FILENAME).write_bytes(manifest)

    with start_client() as client:
        cipher = Cipher(get_transmit_key())
        header = send_request(client, cipher, scope, files)
        return recv_response(client, cipher, header)


if __name__ != "__main__":
    pass
elif not getpass("\nsign package? (Y/n) ").lower().strip().startswith("n"):
    summary = request_package_summary_signature(
        get_transmit_key=current_transmit_key
    )
    PackageVerifier(SIGNING_KEY, path="").verify_summary(summary)
    Path(SIGNATURE_FILENAME).write_bytes(summary)
elif not getpass("verify package? (Y/n) ").lower().strip().startswith("n"):
    summary = Path(SIGNATURE_FILENAME).read_bytes()
    PackageVerifier(SIGNING_KEY, path="").verify_summary(summary)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import io
import random
import warnings

from aiootp import _paths as p

from conftest import *


random.seed(csprng(2500))


KEY = token_bytes(32)
SALT = token_bytes(32)


def select_k_ints(start: int, end: int, k: int) -> t.List[int]:
    """
    Constructs a random distinct list of unique ints for testing a range
    of options that always includes the extremes of the range.
    """
    k = k - 2 if k > 2 else 0
    ints = [*range(start + 1, end - 1)]
    random.shuffle(ints)
    return [start, *sorted(ints[:k]), end - 1]


def is_windows_limitation(_: Ignore) -> bool:
    """
    On Windows, ``os.chmod(path, 0o000)`` doesn't deny reading attempts
    the way Unix-like systems do.

    This function raises the intended PermissionError if reading wasn't
    denied & the tests are being run on an OS which shouldn't have this
    Windows limitation.
    """
    if platform.system() != "Windows":
        raise PermissionError("Set permissions allowed reading.")

    warnings.warn(
        "NOTICE: File permissions on Windows don't deny reading of "
        "sensitive files, like random seed & salt files, as Unix-like "
        "systems do when os.chmod(path, 0o000) is used."
    )


class Targets(t.NamedTuple):
    """
    A container for distinct target test functions.
    """

    asynch: t.Callable[..., t.Awaitable[t.Any]]
    synch: t.Callable[..., t.Any]


class TargetRunner:
    """
    A runner base class for this file's unit tests.
    """

    targets: Targets

    async def run(self, target, *a: t.Any, **kw: t.Any) -> t.Any:
        if target is self.targets.asynch:
            return await target(*a, **kw)
        else:
            return target(*a, **kw)


class TestDeniableFilename(TargetRunner):
    """
    Tests for the hash to shrunken filename functionalities.
    """

    targets = Targets(
        asynch=p.adeniable_filename,
        synch=p.deniable_filename,
    )

    @pytest.mark.parametrize("size", select_k_ints(-16, 33, k=6))
    @pytest.mark.parametrize("target", targets)
    async def test_size_arg_must_be_between_1_and_16_inclusive(
        self, target, size: int
    ) -> None:
        try:
            await self.run(target, KEY, size=size)
        except ValueError:
            assert (size > 16) or (size < 1)
        else:
            assert 16 >= size >= 1

    @pytest.mark.parametrize("size", [4, 8])
    @pytest.mark.parametrize("key_length", select_k_ints(0, 33, k=6))
    @pytest.mark.parametrize("target", targets)
    async def test_key_length_must_be_at_least_double_size_arg(
        self, target, key_length: int, size: int
    ) -> None:
        key = token_bytes(key_length)

        try:
            await self.run(target, key, size=size)
        except ValueError:
            assert key_length < 2 * size
        else:
            assert key_length >= 2 * size

    @pytest.mark.parametrize("size", [4, 8])
    @pytest.mark.parametrize("key_length", select_k_ints(16, 33, k=4))
    @pytest.mark.parametrize("target", targets)
    async def test_same_key_and_size_make_same_name(
        self, target, key_length: int, size: int
    ) -> None:
        key = token_bytes(key_length)

        result_0 = await self.run(target, key, size=size)
        result_1 = await self.run(target, key, size=size)

        assert result_0 == result_1

    @pytest.mark.parametrize("size", select_k_ints(4, 17, k=6))
    @pytest.mark.parametrize("key_length", select_k_ints(48, 65, k=4))
    @pytest.mark.parametrize("target", targets)
    async def test_different_key_or_size_make_different_name(
        self, target, key_length: int, size: int
    ) -> None:
        key = csprng(key_length)

        result_0 = await self.run(target, key, size=size)
        result_1 = await self.run(target, key, size=size - 1)
        result_2 = await self.run(target, key + b"\xff", size=size)

        assert result_0 != result_1
        assert result_0 != result_2
        assert result_1 != result_2

    @pytest.mark.parametrize("size", select_k_ints(2, 17, k=4))
    @pytest.mark.parametrize("key_size_multiple", select_k_ints(2, 8, k=4))
    @pytest.mark.parametrize("target", targets)
    async def test_same_names_when_xor_of_key_sections_are_same(
        self, target, key_size_multiple: int, size: int
    ) -> None:
        raw_control = csprng(size)
        control = int.from_bytes(raw_control, "big")
        results = {await self.run(target, 3 * raw_control, size=size)}

        for _ in range(number_of_trials := 4):
            xor_result = 0
            key = csprng(key_size_multiple * size)
            key_reader = io.BytesIO(key)
            while section := key_reader.read(size):
                xor_result ^= int.from_bytes(section, "big")
            negative = (xor_result ^ control).to_bytes(size, "big")
            results.add(await self.run(target, key + negative, size=size))

        assert len(results) != number_of_trials
        assert len(results) == 1


class TestFindSaltFile(TargetRunner):
    """
    Tests for the finding protected salt file functionalities.
    """

    targets = Targets(
        asynch=p._afind_salt_file,
        synch=p._find_salt_file,
    )

    @pytest.mark.parametrize("target", targets)
    async def test_salt_is_found_in_provided_dir_with_a_deniable_filename(
        self, target
    ) -> None:
        parent = p.SecurePath()
        stem = p.deniable_filename(key=KEY, size=8)
        path = await self.run(target, parent, key=KEY, size=8)

        assert parent == path.parent
        assert stem == str(path.stem)
        assert parent / stem == path


class TestMakeSaltFile(TargetRunner):
    """
    Tests for the making protected salt file functionalities.
    """

    targets = Targets(
        asynch=p._amake_salt_file,
        synch=p._make_salt_file,
    )

    @pytest.mark.parametrize("target", targets)
    async def test_given_salt_is_persisted(
        self, target, salt_path: t.Path
    ) -> None:
        await self.run(target, salt_path, salt=SALT)

        salt_path.chmod(0o600)

        assert SALT == salt_path.read_bytes()

    @pytest.mark.parametrize("target", targets)
    async def test_new_salt_is_created(
        self, target, salt_path: t.Path
    ) -> None:
        await self.run(target, salt_path)

        salt_path.chmod(0o600)

        salt = salt_path.read_bytes()
        assert len(salt) == 32
        assert len(set(salt)) >= 16

    @pytest.mark.parametrize("size", select_k_ints(16, 49, k=4))
    @pytest.mark.parametrize("target", targets)
    async def test_salt_must_be_at_least_32_bytes(
        self, target, salt_path: t.Path, size: int
    ) -> None:
        salt = token_bytes(size)

        try:
            await self.run(target, salt_path, salt=salt)
        except ValueError:
            assert size < 32
        else:
            assert size >= 32

    @pytest.mark.parametrize("target", targets)
    async def test_no_read_permissions_set_after_creation(
        self, target, salt_path: t.Path
    ) -> None:
        await self.run(target, salt_path)

        with Ignore(PermissionError, if_else=is_windows_limitation):
            salt_path.read_bytes()

    @pytest.mark.parametrize("target", targets)
    async def test_no_write_permissions_set_after_creation(
        self, target, salt_path: t.Path
    ) -> None:
        await self.run(target, salt_path)

        problem = (  # fmt: skip
            "Set permissions allowed writing."
        )
        with Ignore(PermissionError, if_else=violation(problem)):
            salt_path.write_bytes(SALT)


class TestReadSaltFile(TargetRunner):
    """
    Tests for the reading protected salt file functionalities.
    """

    targets = Targets(
        asynch=p._aread_salt_file,
        synch=p._read_salt_file,
    )

    @pytest.mark.parametrize("target", targets)
    async def test_read_doesnt_automate_creation(
        self, target, salt_path: t.Path
    ) -> None:
        problem = (  # fmt: skip
            "Reading non-existant salt file didn't raise error."
        )
        with Ignore(FileNotFoundError, if_else=violation(problem)):
            await self.run(target, salt_path)

    @pytest.mark.parametrize("size", select_k_ints(16, 49, k=4))
    @pytest.mark.parametrize("target", targets)
    async def test_salt_must_be_at_least_32_bytes(
        self, target, salt_path: t.Path, size: int
    ) -> None:
        salt_path.write_bytes(token_bytes(size))
        salt_path.chmod(0o000)

        try:
            await self.run(target, salt_path)
        except ValueError:
            assert size < 32
        else:
            assert size >= 32

    @pytest.mark.parametrize("target", targets)
    async def test_no_read_permissions_set_after_reading(
        self, target, salt_path: t.Path
    ) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        await self.run(target, salt_path)

        with Ignore(PermissionError, if_else=is_windows_limitation):
            salt_path.read_bytes()

    @pytest.mark.parametrize("target", targets)
    async def test_no_write_permissions_set_after_reading(
        self, target, salt_path: t.Path
    ) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        await self.run(target, salt_path)

        problem = (  # fmt: skip
            "Set permissions allowed writing."
        )
        with Ignore(PermissionError, if_else=violation(problem)):
            salt_path.write_bytes(SALT)

    @pytest.mark.parametrize("target", targets)
    async def test_retrieves_persisted_salt(
        self, target, salt_path: t.Path
    ) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        assert SALT == await self.run(target, salt_path)


class TestUpdateSaltFile(TargetRunner):
    """
    Tests for the updating protected salt file functionalities.
    """

    targets = Targets(
        asynch=p._aupdate_salt_file,
        synch=p._update_salt_file,
    )

    @pytest.mark.parametrize("target", targets)
    async def test_new_salt_is_created(
        self, target, salt_path: t.Path
    ) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        new_salt = token_bytes(32)
        await self.run(target, salt_path, salt=new_salt)

        salt_path.chmod(0o600)
        assert new_salt == salt_path.read_bytes()

    @pytest.mark.parametrize("size", select_k_ints(16, 49, k=4))
    @pytest.mark.parametrize("target", targets)
    async def test_salt_must_be_at_least_32_bytes(
        self, target, salt_path: t.Path, size: int
    ) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        try:
            await self.run(target, salt_path, salt=csprng(size))
        except ValueError:
            assert size < 32
        else:
            assert size >= 32

    @pytest.mark.parametrize("target", targets)
    async def test_no_read_permissions_set_after_updating(
        self, target, salt_path: t.Path
    ) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        await self.run(target, salt_path, salt=token_bytes(32))

        with Ignore(PermissionError, if_else=is_windows_limitation):
            salt_path.read_bytes()

    @pytest.mark.parametrize("target", targets)
    async def test_no_write_permissions_set_after_updating(
        self, target, salt_path: t.Path
    ) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        await self.run(target, salt_path, salt=token_bytes(32))

        problem = (  # fmt: skip
            "Set permissions allowed writing."
        )
        with Ignore(PermissionError, if_else=violation(problem)):
            salt_path.write_bytes(SALT)


class TestDeleteSaltFile(TargetRunner):
    """
    Tests for the deleting protected salt file functionalities.
    """

    targets = Targets(
        asynch=p._adelete_salt_file,
        synch=p._delete_salt_file,
    )

    @pytest.mark.parametrize("target", targets)
    async def test_file_is_removed(self, target, salt_path: t.Path) -> None:
        salt_path.write_bytes(SALT)
        salt_path.chmod(0o000)

        await self.run(target, salt_path)

        assert not salt_path.is_file()
        assert not salt_path.is_dir()


class AdminContext(dict):
    """
    A container for admin data needed by the test target.
    """

    path: t.Path = p.AdminPath()


class UserContext(dict):
    """
    A container for user data needed by the test target.
    """

    path: t.Path = p.SecurePath()


class TargetContexts(t.NamedTuple):
    """
    A container for distinct data contexts needed by the test target.
    """

    admin_ctx: AdminContext
    user_ctx: UserContext


class TestZZZSecureSaltPath(TargetRunner):
    """
    Tests for the protected salt file path interface functionalities.
    """

    kwargs = TargetContexts(
        admin_ctx=AdminContext(key=KEY, _admin=True),
        user_ctx=UserContext(key=KEY),
    )

    targets = Targets(
        asynch=p.AsyncSecureSaltPath,
        synch=p.SecureSaltPath,
    )

    @pytest.mark.parametrize("kw", kwargs)
    @pytest.mark.parametrize("target", targets)
    async def test_default_secure_path(self, target, kw) -> None:
        result = await self.run(target, **kw)

        assert isinstance(result, t.Path)
        assert str(kw.path) == str(result.parent)

    @pytest.mark.parametrize("kw", kwargs)
    @pytest.mark.parametrize("target", targets)
    async def test_same_keys_derive_same_paths(self, target, kw) -> None:
        result_0 = await self.run(target, **kw)
        result_1 = await self.run(target, **kw)

        assert isinstance(result_0, t.Path)
        assert isinstance(result_1, t.Path)
        assert str(result_0) == str(result_1)

    @pytest.mark.parametrize("kw", kwargs)
    @pytest.mark.parametrize("target", targets)
    async def test_salt_is_static(self, target, kw) -> None:
        path_0 = await self.run(target, **kw)
        result_0 = p._read_salt_file(path_0)

        path_1 = await self.run(target, **kw)
        result_1 = p._read_salt_file(path_1)

        assert len(result_0) == 32
        assert len(result_1) == 32
        assert isinstance(result_0, bytes)
        assert isinstance(result_1, bytes)
        assert result_0 == result_1

    @pytest.mark.parametrize("kw", kwargs)
    @pytest.mark.parametrize("target", targets)
    async def test_first_time_calls_create_salt(self, target, kw) -> None:
        path = await self.run(target, **kw)

        result_0 = p._read_salt_file(path)
        p._delete_salt_file(path)

        await self.run(target, **kw)
        result_1 = p._read_salt_file(path)

        assert len(result_0) == 32
        assert len(result_1) == 32
        assert isinstance(result_0, bytes)
        assert isinstance(result_1, bytes)
        assert result_0 != result_1

    @pytest.mark.parametrize("kw", kwargs)
    @pytest.mark.parametrize("target", targets)
    async def test_zzz_remove_test_files(self, target, kw) -> None:
        path = await self.run(target, **kw)

        if path.is_file():
            path.chmod(0o600)
            path.unlink()


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A multi-processing interface.
"""

__all__ = ["Processes", "get_process_id"]


import multiprocessing
from multiprocessing.context import SpawnContext
from os import getpid as get_process_id
from concurrent.futures import process
from concurrent.futures import ProcessPoolExecutor

from aiootp._typing import Typing as t

from .concurrency_interface import ConcurrencyInterface


class Processes(ConcurrencyInterface):
    """
    Simplifies spawning & returning the values produced by `Process` &
    `ProcessPoolExecutor` objects with an interface over these types
    from the `multiprocessing` & `concurrent.futures` packages.
    """

    __slots__ = ()

    _default_probe_delay: float = 0.005
    _context: SpawnContext = multiprocessing.get_context("spawn")
    _pool: t.PoolExecutorType = ProcessPoolExecutor(mp_context=_context)
    _type: type = _context.Process

    BrokenPool: type = process.BrokenProcessPool

    get_id: t.Callable[[], int] = get_process_id

    @classmethod
    def _get_queue(cls, /, maxsize: int = 1) -> t.QueueType:
        """
        Returns a queue object to retrieve values from spawned workers
        with the class' default multiprocessing context.
        """
        return cls._context.Queue(maxsize=maxsize)

    @classmethod
    def reset_pool(cls, /) -> None:
        """
        When a process pool is broken by being abruptly exited, this
        method can be called to reset the class' pool object with a new
        instance with its default multiprocessing context.
        """
        cls._pool = cls._pool.__class__(mp_context=cls._context)


module_api = dict(
    Processes=t.add_type(Processes),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    get_process_id=get_process_id,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2025 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from aiootp.commons.namespaces import OpenNamespace


chunky2048_test_vector_0 = OpenNamespace(
    plaintext=(
        b"it is not a doctrine to be preached, but a deed to be done."
    ),
    key=bytes.fromhex(
        "54868dc506d99611adb67f1b41eda44fd7151e135860a6d791cad6df6715bd"
        "5204713bc2b064e88a429b93aaa282d9cc7c0c5cdaaffb65cc268b7acedd5e"
        "531a"
    ),
    shmac=bytes.fromhex(
        "f791571464dd45754e16d8e6dea2f1a05204dca2510589be012bc5117dfd98e8"
    ),
    salt=bytes.fromhex("cf3761d57860bdbc"),
    iv=bytes.fromhex("e09d2eee03ae82e6"),
    aad=b"test_vector_I",
    ciphertext=bytes.fromhex(
        "1ce9033362f402390f872fa4f20de14d8a13f32f3906dd302a275b2c62f9f5"
        "f1a8ddb61871a2bab61dffaa2b414c2ca68601a61d9670a944619fc5105469"
        "b95e63c494326678e865a751d2ab0cac40439aa51cec3a8aa9e4cf0255ea96"
        "659a4cce41e2c8dac09f97d5bb05efb39bcf6043c6cd18364f0ecc0c721160"
        "e8204e333b1e336c55bd9a09ba1c0ce08805cc924c1925ad238b785ad17e6c"
        "064b2fe8676dc2fff3a3a895381cbf8d5aa04970c4f210ae323a5274c8f703"
        "4167ba0f7042e17d748d68327d2219c83bf4e4fb9f5d36c0b0c5ca661db09c"
        "b887d2bd39589d2e12f7a799b2e3e1cad7a7dbfd4c9d1a46de6ed5e51d7561"
        "70de526f4c4fa9fb"
    ),
)
